#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "neural_net_data.h"
#include "neural_net_network.h"
#include "neural_net_nparams.h"

#define SIGN0(a)  (((a)==0)?0:(((a)<0)?-1:1))
#define MIN(a,b) (((a)<(b))?(a):(b))
#define MAX(a,b) (((a)>(b))?(a):(b))

int debug = 0;

//TODO: extend(+1layer)

enum act_type {act_rlu, act_softmax, act_lin, act_sigm};
enum upd_type {upd_sgd, upd_rprop};

double RANDW = 0.1;
double eta = 0.1, etaC = 0.0;
double mu = 0.5;
double minDG = 0.01;
double wdecay = 0.0;

double DELTAMIN = 1e-6;
double DELTAMAX = 10.0;
double ETAPLUS = 1.2;
double ETAMINUS = 0.5;

//double WMAX = 1.0;
double WMAX = 10000000.0;

int updateRule = 0;

double sumv = 0, sumv0 = 0, sumt = 0;

int updateT = 1;

void fprintact(FILE* f, struct network* net, int* ptn); 
void fprintgrad(FILE* f, struct network* net); 
void fprintw(FILE* f, struct network* net); 

void clearw(double* w, int n) {
  int i;
  for (i = 0; i < n; i++) *w++ = 0;
}

void setw(double* w, int n, double v) {
  int i;
  for (i = 0; i < n; i++) *w++ = v;
}

struct layer/* {
  int N;
  int fieldsz;
  int shared;
  int act;
  int wsz;
  int useBias;
  int useResidual;
  double *weights;
  double *dw, *delta, *deltaw, *dy;
  int *widx, *bidx, *rwidx;
  int **inputidx;
  double* out0;
  double* out;
  int gn;
}*/;

struct network/* {
  int L;
  struct layer* layers;
}*/;

int initwidx(struct network* net, int l) {
  int m = net->layers[l].fieldsz;
  if (l > 0) m *= net->layers[l-1].N;
  else m *= featN;
  if (net->layers[l].useBias) net->layers[l].bidx = malloc(PTNSZ2 * net->layers[l].N * sizeof(int));
  else net->layers[l].bidx = NULL;
  net->layers[l].widx = malloc(PTNSZ2 * net->layers[l].N * sizeof(int));
  fprintf(stderr, "malloc %d %p\n", (int) (PTNSZ2 * net->layers[l].N * sizeof(int)), net->layers[l].widx);
  int i, j, ij = 0; 
  int sz = 0, wsz = 0;
  for (i = 0; i < PTNSZ2; i++) {
    for (j = 0; j < net->layers[l].N; j++) {
      net->layers[l].widx[ij++] = sz;
      sz += m;
      if ((i == 0) || (!net->layers[l].shared)) wsz += m; 
    }
    if (net->layers[l].shared) sz = 0;
  }
  if (net->layers[l].useBias) {
    sz = wsz;
    ij = 0;
    for (i = 0; i < PTNSZ2; i++)
      for (j = 0; j < net->layers[l].N; j++) {
	net->layers[l].bidx[ij++] = sz;
	sz++;
	wsz++;
      }
  }

  if (net->layers[l].shared) return wsz;
  else return wsz;
}

int initrwidx(struct network* net, int l, int wsz0) {
  int m = net->layers[l].fieldsz;
  if (l > 1) m *= net->layers[l-2].N;
  else m *= featN;
  net->layers[l].rwidx = malloc(PTNSZ2 * net->layers[l].N * sizeof(int));
  fprintf(stderr, "malloc %d %p\n", (int) (PTNSZ2 * net->layers[l].N * sizeof(int)), net->layers[l].rwidx);
  int i, j, ij = 0; 
  int sz = 0, wsz = wsz0;
  for (i = 0; i < PTNSZ2; i++) {
    for (j = 0; j < net->layers[l].N; j++) {
      net->layers[l].rwidx[ij++] = sz;
      sz += m;
      if ((i == 0) || (!net->layers[l].shared)) wsz += m; 
    }
    if (net->layers[l].shared) sz = 0;
  }

  if (net->layers[l].shared) return wsz;
  else return wsz;
}


void initiidx(struct network* net, int l) {
  net->layers[l].inputidx = malloc(PTNSZ2*sizeof(int*));
  int i;
  for (i = 0; i < PTNSZ2; i++) {
    net->layers[l].inputidx[i] = malloc(net->layers[l].fieldsz*sizeof(int));
    int x, y, x1, y1, ii = 0;
    int fsz = ((int) sqrt(net->layers[l].fieldsz))/2;
    x = i / PTNSZ;
    y = i % PTNSZ;
    for (x1 = x - fsz; x1 <= x + fsz; x1++)
      for (y1 = y - fsz; y1 <= y + fsz; y1++) 
	if ((x1 < 0) || (x1 >= PTNSZ) || (y1 < 0) || (y1 >= PTNSZ)) 
	  net->layers[l].inputidx[i][ii++] = -1; 
	else
	  net->layers[l].inputidx[i][ii++] = x1*PTNSZ + y1; 
  }
}


void randw(struct network* net, int l) {
  int j;
  if (debug) 
    for (j = 0; j < net->layers[l].wsz; j++)
      net->layers[l].weights[j] = 1;
  else
    for (j = 0; j < net->layers[l].wsz; j++)
      net->layers[l].weights[j] = 2.0*RANDW * rand()/(RAND_MAX+1.0) - RANDW;
}

// this should be improved...
int getVal(FILE* f, int l, char* name) {
  char s[1000];
  int v;
  if (l == -1) sprintf(s, "%s=%%d", name);
  else sprintf(s, "%d:%s=%%d", l, name); 
  char buff[1000];
  if (!fgets(buff, 1000, f)) perror("");
  sscanf(buff, s, &v);
  //  fprintf(stderr, "reading net: %s from %d\n", s, buff);
  return v;
}

void printVal(FILE* f, int l, char* name, int v) {
  if (l == -1) fprintf(f, "%s=%d\n", name, v);
  else fprintf(f, "%d:%s=%d\n", l, name, v);
}

struct network* create_net(char* netfile, int train) {
  int l;
  struct network* net = malloc(sizeof(struct network));

  struct nparams* npar = init_nparams(netfile);
  net->L = getVal_int(npar, -1, "L", 1);
  net->layers = malloc(net->L * sizeof(struct layer));
  for (l = 0; l < net->L; l++) {
    net->layers[l].N = getVal_int(npar, l, "N", 1);
    net->layers[l].fieldsz = getVal_int(npar, l, "fieldsz", 1);
    net->layers[l].shared = getVal_int(npar, l, "shared", 0);
    net->layers[l].act = getVal_int(npar, l, "act", 0);
    net->layers[l].wsz = getVal_int(npar, l, "wsz", -1);
    net->layers[l].useBias = getVal_int(npar, l, "bias", 0);
    net->layers[l].useResidual = getVal_int(npar, l, "residual", 0);

    fprintf(stderr, "net: %d: N=%d fieldsz=%d shared=%d act=%d wsz=%d bias=%d residual=%d\n", l, 
	    net->layers[l].N, net->layers[l].fieldsz, net->layers[l].shared,
	    net->layers[l].act, net->layers[l].wsz, net->layers[l].useBias, net->layers[l].useResidual);
  }
  close_nparams(npar);

  for (l = 0; l < net->L; l++) {
    int wsz = initwidx(net, l);
    if (net->layers[l].useResidual) wsz = initrwidx(net, l, wsz);
    else net->layers[l].rwidx = NULL;
    if (net->layers[l].wsz == -1) {
      fprintf(stderr, "net %d:wsz set to %d\n", l, wsz);
      net->layers[l].wsz = wsz;
      net->layers[l].weights = malloc(net->layers[l].wsz * sizeof(double));
      randw(net, l);
      fprintf(stderr, "net: %d: allocated %d\n", l, (int) (net->layers[l].wsz * sizeof(double)));
    }
    else {
      net->layers[l].weights = malloc(net->layers[l].wsz * sizeof(double));
      char s[1000];
      sprintf(s, "%s.%d", netfile, l);
      FILE* f = fopen(s, "r");
      if (fread(net->layers[l].weights, net->layers[l].wsz*sizeof(double), 1, f) != 1) perror("");
      fclose(f);
    }

    net->layers[l].out = malloc(net->layers[l].N * PTNSZ2 * sizeof(double));
    net->layers[l].out0 = malloc(net->layers[l].N * PTNSZ2 * sizeof(double));

    initiidx(net, l);

    if (train) {
      net->layers[l].dw = malloc(net->layers[l].wsz * sizeof(double));
      net->layers[l].deltaw = malloc(net->layers[l].wsz * sizeof(double));
      if (updateRule == upd_rprop) net->layers[l].delta = malloc(net->layers[l].wsz * sizeof(double));
      net->layers[l].dy = malloc(net->layers[l].N * PTNSZ2 * sizeof(double));
      clearw(net->layers[l].dw, net->layers[l].wsz);
      clearw(net->layers[l].deltaw, net->layers[l].wsz);
      if (updateRule == upd_rprop) setw(net->layers[l].delta, net->layers[l].wsz, eta);
      clearw(net->layers[l].dy, net->layers[l].N * PTNSZ2);
      net->layers[l].gn = 0;
      fprintf(stderr, "net: train allocated\n");
    }
    else 
      net->layers[l].dw = net->layers[l].deltaw = net->layers[l].dy = NULL;

  }
  return net;
}
void print_net(char* netfile, struct network* net) {
  char s[1000];
  FILE* f = fopen(netfile, "w");
  int l;
  printVal(f, -1, "L", net->L);
  for (l = 0; l < net->L; l++) {
    printVal(f, l, "N", net->layers[l].N);
    printVal(f, l, "fieldsz", net->layers[l].fieldsz);
    printVal(f, l, "shared", net->layers[l].shared);
    printVal(f, l, "act", net->layers[l].act);
    printVal(f, l, "wsz", net->layers[l].wsz);
    printVal(f, l, "bias", net->layers[l].useBias);
    printVal(f, l, "residual", net->layers[l].useResidual);
  }
  fclose(f);
  for (l = 0; l < net->L; l++) {
    sprintf(s, "%s.%d", netfile, l);
    f = fopen(s, "w");
    fwrite(net->layers[l].weights, net->layers[l].wsz*sizeof(double), 1, f);
    fclose(f);
    if (net->layers[l].deltaw != NULL) {
      sprintf(s, "%s.dw%d", netfile, l);
      f = fopen(s, "w");
      fwrite(net->layers[l].deltaw, net->layers[l].wsz*sizeof(double), 1, f);
      fclose(f);
    }
    if (net->layers[l].delta != NULL) {
      sprintf(s, "%s.d%d", netfile, l);
      f = fopen(s, "w");
      fwrite(net->layers[l].delta, net->layers[l].wsz*sizeof(double), 1, f);
      fclose(f);
    }
  }
}
void del_net(struct network* net) {
  int l;
  for (l = 0; l < net->L; l++) {
    free(net->layers[l].weights);
    free(net->layers[l].widx);
    if (net->layers[l].bidx) free(net->layers[l].bidx);
    if (net->layers[l].rwidx) free(net->layers[l].rwidx);
    free(net->layers[l].out0);
    free(net->layers[l].out);
    int i;
    for (i = 0; i < net->layers[l].fieldsz; i++)
      free(net->layers[l].inputidx[i]);
    free(net->layers[l].inputidx);
    if (net->layers[l].dw) free(net->layers[l].dw);
    if (net->layers[l].deltaw) free(net->layers[l].deltaw);
    if (net->layers[l].dy) free(net->layers[l].dy);
  }
  free(net->layers);
}

void setinput(struct network* net, int* ptn, int l, int i, double* input) {
  int j, ii1 = 0, N1, k;
  if (l > 0) N1 = net->layers[l-1].N;
  else N1 = featN; 
  for (j = 0; j < net->layers[l].fieldsz; j++) {
    if (net->layers[l].inputidx[i][j] == -1)
      for (k = 0; k < N1; k++)
	input[ii1++] = 0;
    else if (l > 0) {
      double *outp = net->layers[l-1].out + N1 * net->layers[l].inputidx[i][j];
      for (k = 0; k < N1; k++)
	input[ii1++] = *outp++;
    }
    else {
      int v = (ptn[net->layers[l].inputidx[i][j]] >> 2);
      for (k = 0; k < N1; k++, v = v >> 1)
	input[ii1++] = (v & 1);
    }
  }
}


void setrinput(struct network* net, int* ptn, int l, int i, double* input) {
  int j, ii1 = 0, N1, k;
  if (l > 1) N1 = net->layers[l-2].N;
  else N1 = featN; 
  for (j = 0; j < net->layers[l].fieldsz; j++) {
    if (net->layers[l].inputidx[i][j] == -1)
      for (k = 0; k < N1; k++)
	input[ii1++] = 0;
    else if (l > 1) {
      double *outp = net->layers[l-2].out + N1 * net->layers[l].inputidx[i][j];
      for (k = 0; k < N1; k++)
	input[ii1++] = *outp++;
    }
    else {
      int v = (ptn[net->layers[l].inputidx[i][j]] >> 2);
      for (k = 0; k < N1; k++, v = v >> 1)
	input[ii1++] = (v & 1);
    }
  }
}


void adddy(struct network* net, int l, int i, double* dy) {
  int j, ii1 = 0, N1 = net->layers[l-1].N, k;
  for (j = 0; j < net->layers[l].fieldsz; j++) {
    if (net->layers[l].inputidx[i][j] != -1) {
      double *dp = net->layers[l-1].dy + N1 * net->layers[l].inputidx[i][j];
      for (k = 0; k < N1; k++)
	*dp++ += dy[ii1++];
    }
  }
}
void addrdy(struct network* net, int l, int i, double* dy) {
  int j, ii1 = 0, N1 = net->layers[l-2].N, k;
  for (j = 0; j < net->layers[l].fieldsz; j++) {
    if (net->layers[l].inputidx[i][j] != -1) {
      double *dp = net->layers[l-2].dy + N1 * net->layers[l].inputidx[i][j];
      for (k = 0; k < N1; k++)
	*dp++ += dy[ii1++];
    }
  }
}

void responsel(struct network* net, int* ptn, int l) {
  int M, rM = 0;
  if (l > 0) M = net->layers[l-1].N;
  else M = featN;
  M *= net->layers[l].fieldsz;
  
  if (net->layers[l].useResidual) {
    if (l > 1) rM = net->layers[l-2].N;
    else rM = featN;
    rM *= net->layers[l].fieldsz;
  }

  int i, ij = 0;
  int n = net->layers[l].N;
  for (i = 0; i < PTNSZ2; i++) {
    double input[M];
    double rinput[rM];
    setinput(net, ptn, l, i, input);
    if (net->layers[l].useResidual) setrinput(net, ptn, l, i, rinput);
    int j;
    for (j = 0; j < n; j++) {
      double v = 0; 
      if (net->layers[l].useBias) v = net->layers[l].weights[net->layers[l].bidx[ij]];
      double *w = net->layers[l].weights + net->layers[l].widx[ij];
      int k;
      for (k = 0; k < M; k++)
	v += input[k] * (*w++);
      if (net->layers[l].useResidual) {
	double *rw = net->layers[l].weights + net->layers[l].rwidx[ij];
	for (k = 0; k < rM; k++)
	  v += rinput[k] * (*rw++);
      }
      
      net->layers[l].out0[i*n + j] = v;
      ij++;
    }
  }

  int Nl = PTNSZ2*net->layers[l].N;
  if (net->layers[l].act == act_softmax) {
    double vmax = net->layers[l].out0[0];
    double vsum = 0;
    for (i = 0; i < Nl; i++)
      if ((!(ptn[i] & 2) || (l < net->L-1)) && 
	  (net->layers[l].out0[i] > vmax)) vmax = net->layers[l].out0[i];
    for (i = 0; i < Nl; i++) {
      net->layers[l].out[i] = exp(net->layers[l].out0[i] - vmax);
      if (!(ptn[i] & 2) || (l < net->L-1)) vsum += net->layers[l].out[i];
    }
    for (i = 0; i < Nl; i++) 
      net->layers[l].out[i] /= vsum;
  }
  else
    for (i = 0; i < Nl; i++) {
      if (net->layers[l].act == act_rlu) {
	if (net->layers[l].out0[i] > 0)
	  net->layers[l].out[i] = net->layers[l].out0[i];
	else
	  net->layers[l].out[i] = 0;
      }
      else if (net->layers[l].act == act_sigm) 
	net->layers[l].out[i] = 1.0/(1.0+exp(-net->layers[l].out0[i])); 
      else
	net->layers[l].out[i] = net->layers[l].out0[i];
    }
  
  for (i = 0; i < Nl; i++) {
    sumv += net->layers[l].out[i]* net->layers[l].out[i];
    sumv0 += net->layers[l].out0[i]* net->layers[l].out0[i];
    sumt++;
  }
}

double* response(struct network* net, int* ptn) {
  int l;
  for (l = 0; l < net->L; l++)
    responsel(net, ptn, l);

  return net->layers[net->L - 1].out;
}

void computegradientl(struct network* net, int* ptn, int l) {
  int i;
  int Nl = PTNSZ2*net->layers[l].N;
  for (i = 0; i < Nl; i++) {
    if (net->layers[l].act == act_rlu) {
      if (net->layers[l].out0[i] == 0)
	net->layers[l].dy[i] = 0;
    }
    else if (net->layers[l].act == act_sigm) {
      net->layers[l].dy[i] *= net->layers[l].out[i]*(1.0 - net->layers[l].out[i]) + minDG/2;
    }
  }

  //  if (l > 0) clearw(net->layers[l-1].dy, net->layers[l-1].N * PTNSZ2);
  
  int M, rM = 0;
  if (l > 0) M = net->layers[l-1].N;
  else M = featN;
  M *= net->layers[l].fieldsz;

  if (net->layers[l].useResidual) {
    if (l > 1) rM = net->layers[l-2].N;
    else rM = featN;
    rM *= net->layers[l].fieldsz;
  }


  int ij = 0;
  int n = net->layers[l].N;

  for (i = 0; i < PTNSZ2; i++) {
    double input[M], dh[M], rinput[rM], rdh[rM];
    setinput(net, ptn, l, i, input);
    if (net->layers[l].useResidual) setrinput(net, ptn, l, i, rinput);
    if (l > 0) {
      int k;
      for (k = 0; k < M; k++) dh[k] = 0;
    }
    if ((l > 1) && (net->layers[l].useResidual)) {
      int k;
      for (k = 0; k < rM; k++) rdh[k] = 0;
    }

    int j;
    for (j = 0; j < n; j++) {
      double *dw = net->layers[l].dw + net->layers[l].widx[ij];
      double *w = net->layers[l].weights + net->layers[l].widx[ij];
      double dy = net->layers[l].dy[ij];

      if (net->layers[l].useBias) net->layers[l].dw[net->layers[l].bidx[ij]] += dy;

      int k;

      if (l > 0) 
	for (k = 0; k < M; k++) { 
	  (*dw++) += input[k] * dy;
	  dh[k] += (*w++) * dy;
	}
      else
	for (k = 0; k < M; k++) 
	  (*dw++) += input[k] * dy;

      if (net->layers[l].useResidual) {
	double *rw = net->layers[l].weights + net->layers[l].rwidx[ij];
	double *rdw = net->layers[l].dw + net->layers[l].rwidx[ij];
	if (l > 1) 
	  for (k = 0; k < rM; k++) { 
	    (*rdw++) += input[k] * dy;
	    rdh[k] += (*rw++) * dy;
	  }
	else
	  for (k = 0; k < rM; k++) 
	    (*rdw++) += rinput[k] * dy;
      }

      ij++;
    }

    if (l > 0) adddy(net, l, i, dh);
    if ((l > 1) && (net->layers[l].useResidual)) addrdy(net, l, i, dh);
  }

  net->layers[l].gn++;
}

void computegradient(struct network* net, int* input, double* target, int* valid, int resp) {
  // if not preceeded by response with same input
  if (resp) response(net, input);

  double diff = 0;
  int i;
  for (i = 0; i < PTNSZ2; i++) {
    if (valid[i] == 0) net->layers[net->L-1].dy[i] = 0;
    else net->layers[net->L-1].dy[i] = (target[i] - net->layers[net->L-1].out[i]);
    diff += net->layers[net->L-1].dy[i];
  }
  //  fprintf(stderr, "diff %f\n", diff);
  
  int l;

  for (l = net->L-2; l >= 0; l--)
    clearw(net->layers[l].dy, net->layers[l].N * PTNSZ2);

  for (l = net->L-1; l >= 0; l--)
    computegradientl(net, input, l);

  if (debug) {
    fprintact(stderr, net, input);
    if (debug >= 2) fprintgrad(stderr, net);
    if (debug >= 3) fprintw(stderr, net);
  }
}


void update(struct network* net) {
  int l;
  for (l = 0; l < net->L; l++) {
    //    if ((net->layers[l].N == 1) && (net->layers[l].fieldsz == 1)) ;
    //    else
    if (net->layers[l].gn) {
      int i = 0;
      double *w, *dw, *deltaw;
      double etal = eta/pow(updateT, etaC)/net->layers[l].gn;
      if (updateRule == upd_rprop) {
	double *delta;
	for (i = 0, w = net->layers[l].weights, dw = net->layers[l].dw, deltaw = net->layers[l].deltaw,
	       delta = net->layers[l].delta;
	     i < net->layers[l].wsz; i++, w++, dw++, deltaw++, delta++) {
	  *dw -= wdecay * *w;
	  double direction = *dw * *deltaw;
	  if (direction > 0) {
	    *delta = MIN(*delta * ETAPLUS, DELTAMAX);
	    *deltaw = SIGN0(*dw) * *delta;
	  }
	  else if (direction < 0) {
	    *delta = MAX(*delta * ETAMINUS, DELTAMIN);
	    *deltaw = 0;
	  }
	  else {
	    *deltaw = SIGN0(*dw) * *delta;
	  }
	  if (*dw != 0) {
	    *w += *deltaw; 
	    if (*w > WMAX) *w = WMAX;
	    if (*w < -WMAX) *w = -WMAX;
	    *dw = 0;
	  }
	}
      }
      else
	for (i = 0, w = net->layers[l].weights, dw = net->layers[l].dw, deltaw = net->layers[l].deltaw;
	     i < net->layers[l].wsz; i++, w++, dw++, deltaw++) {
	  *deltaw = etal * (*dw - wdecay * net->layers[l].gn * *w) + mu * (*deltaw);
	  *w += *deltaw;
	  *dw = 0;
	}
      net->layers[l].gn = 0;
    }
  }

  fprintf(stderr, "sumv %f %f\n", sumv/sumt, sumv0/sumt);
  sumv = sumv0 = sumt = 0;
  fprintf(stderr, "eta: %f\n", eta/pow(updateT, etaC));
  updateT++;
}

double* wsize(struct network* net) {
  double *sz = malloc(sizeof(double)*net->L);
  int l;
  for (l = 0; l < net->L; l++) {
    double *w;
    int i;
    sz[l] = 0;
    for (i = 0, w = net->layers[l].weights; i < net->layers[l].wsz; i++, w++)
      sz[l] += (*w)*(*w);
    sz[l] /= net->layers[l].wsz;
  }
  return sz;
}
double* dwsize(struct network* net) {
  double *sz = malloc(sizeof(double)*net->L);
  int l;
  for (l = 0; l < net->L; l++) {
    double *dw;
    int i;
    sz[l] = 0;
    for (i = 0, dw = net->layers[l].dw; i < net->layers[l].wsz; i++, dw++)
      sz[l] += (*dw)*(*dw);
    sz[l] /= ((double) net->layers[l].gn * net->layers[l].gn * net->layers[l].wsz);
  }
  return sz;
}

int layersN(struct network* net) {
  return net->L;
}

void fprintact(FILE* f, struct network* net, int* ptn) {
  int i, j, l;
  for (i = 0; i < PTNSZ2; i++) {
    fprintf(f, "%03x ", ptn[i]);
    if (i % PTNSZ == PTNSZ-1) fprintf(f, "\n");
  }
  
  for (l = 0; l < net->L; l++) {
    fprintf(f, "layer %d\n", l);
    int n = net->layers[l].N;
    double sum;
    for (j = 0; j < n; j++) {
      sum = 0;
      for (i = 0; i < PTNSZ2; i++) {
	fprintf(f, "%f ", net->layers[l].out0[i*n + j]);
	if (i % PTNSZ == PTNSZ-1) fprintf(f, "\n");
      }
      fprintf(f, "out0\n");
      for (i = 0; i < PTNSZ2; i++) {
	fprintf(f, "%f ", net->layers[l].out[i*n + j]);
	if (i % PTNSZ == PTNSZ-1) fprintf(f, "\n");
	sum += net->layers[l].out[i*n+j];
      }
      fprintf(f, "out sum=%f\n", sum);
    }
  }
}


void fprintgrad(FILE* f, struct network* net) {
  int i, j, l;
  for (l = 0; l < net->L; l++) {
    fprintf(f, "grad:layer %d\n", l);
    int n = net->layers[l].N;
    for (j = 0; j < n; j++) {
      for (i = 0; i < PTNSZ2; i++) {
	fprintf(f, "%f ", net->layers[l].dy[i*n + j]);
	if (i % PTNSZ == PTNSZ-1) fprintf(f, "\n");
      }
      fprintf(f, "dy\n");
    }
  }
}

void fprintw(FILE* f, struct network* net) {
  int i, l;
  for (l = 0; l < net->L; l++) {
    int m = net->layers[l].fieldsz;
    if (l > 0) m *= net->layers[l-1].N;
    else m *= featN;

    fprintf(f, "w:layer %d\n", l);
    for (i = 0; i < net->layers[l].wsz; i++) {
      fprintf(f, "%f ", net->layers[l].weights[i]);
      if (i%m == m-1) fprintf(stderr, "\n");
    }
  }
}


void unshare(struct network* net, int l) {
  if (net->layers[l].shared == 0) return;

  struct layer oldlayer = net->layers[l];
  net->layers[l].shared = 0;
  int wsz = initwidx(net, l);
  if (net->layers[l].useResidual) wsz = initrwidx(net, l, wsz);
  else net->layers[l].rwidx = NULL;

  net->layers[l].wsz = wsz;
  net->layers[l].weights = malloc(net->layers[l].wsz * sizeof(double));

  net->layers[l].dw = malloc(net->layers[l].wsz * sizeof(double));
  net->layers[l].deltaw = malloc(net->layers[l].wsz * sizeof(double));
  if (updateRule == upd_rprop) net->layers[l].delta = malloc(net->layers[l].wsz * sizeof(double));
  clearw(net->layers[l].dw, net->layers[l].wsz);
  clearw(net->layers[l].deltaw, net->layers[l].wsz);
  if (updateRule == upd_rprop) setw(net->layers[l].delta, net->layers[l].wsz, eta);

  int M, rM = 0;
  if (l > 0) M = net->layers[l-1].N;
  else M = featN;
  M *= net->layers[l].fieldsz;
  
  if (net->layers[l].useResidual) {
    if (l > 1) rM = net->layers[l-2].N;
    else rM = featN;
    rM *= net->layers[l].fieldsz;
  }
  int i, ij = 0;
  int n = net->layers[l].N;
  for (i = 0; i < PTNSZ2; i++) {
    int j;
    for (j = 0; j < n; j++) {
      if (net->layers[l].useBias) 
	net->layers[l].weights[net->layers[l].bidx[ij]] = oldlayer.weights[oldlayer.bidx[ij]];
      double *w = net->layers[l].weights + net->layers[l].widx[ij];
      double *ow = oldlayer.weights + oldlayer.widx[ij];
      int k;
      for (k = 0; k < M; k++)
	*w++ = *ow++;
      if (net->layers[l].useResidual) {
	double *rw = net->layers[l].weights + net->layers[l].rwidx[ij];
	double *orw = oldlayer.weights + oldlayer.rwidx[ij];
	for (k = 0; k < rM; k++)
	  *rw++ = *orw++;
      }
      ij++;
    }
  }

  free(oldlayer.weights);
  free(oldlayer.widx);
  if (oldlayer.bidx) free(oldlayer.bidx);
  if (oldlayer.rwidx) free(oldlayer.rwidx);
  if (oldlayer.dw) free(oldlayer.dw);
  if (oldlayer.deltaw) free(oldlayer.deltaw);
}

void unshareTop(struct network* net) {
  unshare(net, net->L-1);
}
