#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "neural_net_data.h"

int featN;
long int ptnSz = 0;

struct data {
  int N, crt, next;
  int *ptns;
  char *ptndir, *ptnfile;
  FILE* f;
  int rewind;
};

void rotate_board(int* ptn, int sym);

int load(struct data* dt, FILE* f) {
  int game, move, sym = 0;
  char str[1000];
  if (!fgets(str, 1000, f)) return 0;
  if (sscanf(str, "%d%d%d", &game, &move, &sym) != 3) {  
    if (sscanf(str, "%d%d", &game, &move) != 2) return 0;
    sym = 0;
  }
  char gamestr[1000];
  sprintf(gamestr, "%s/%d.ptn", dt->ptndir, game);
  FILE* fg = fopen(gamestr, "r");
  if (fg == NULL) {
    char serr[10000];
    sprintf(serr, "loading %s", gamestr);
    perror(serr);
    exit(1);
  }
  if (ptnSz == 0) {
    char buff[100000];
    if (fgets(buff, 100000, fg) == 0) {
      fprintf(stderr, "empty file %s\n", gamestr);
      exit(1);
    }
    ptnSz = ftell(fg);
  }
  fseek(fg, ptnSz*move, SEEK_SET);
  
  int i, *p, ret;
  for (i = 0, p = dt->ptns+dt->next; i < PTNSZ2; i++, p++) 
    if ((ret = fscanf(fg, "%x", p)) != 1) {
      fprintf(stderr, "ret = %d %s %d\n", ret, gamestr, move);
      perror("reading ptns");
      exit(1);
    }
  fclose(fg);
  rotate_board(dt->ptns+dt->next, sym);

  return 1;
}

struct data* create_data(char* ptnfile, char* ptndir, int preload) {
  struct data* dt = malloc(sizeof(struct data));
  dt->N = -1;
  dt->ptns = NULL;
  dt->crt = dt->next = 0;
  dt->ptndir = strdup(ptndir);
  dt->ptnfile = strdup(ptnfile);
  
  if (preload == 1) {
    // allocate memory first
    dt->N = ptnN(dt);
    dt->ptns = malloc(sizeof(int)*PTNSZ2*dt->N);

    // load patterns
    FILE* f = fopen(ptnfile, "r");
    while (load(dt, f)) dt->next += PTNSZ2;
    fclose(f);
  }
  else {
    // allocate only 1
    dt->ptns = malloc(sizeof(int)*PTNSZ2);
    dt->f = fopen(ptnfile, "r");
    dt->rewind = (preload == -1);
  }
  return dt;
}

int* nextptn(struct data* dt) {
  if (dt->next == 0) {
    if (load(dt, dt->f) == 0) {
      if (dt->rewind) {
	//	fprintf(stderr, "rewind\n");
	fseek(dt->f, 0, SEEK_SET);
	if (load(dt, dt->f))
	  return dt->ptns;
      }
      return 0; 
    }
    else return dt->ptns;
  }
  if (dt->crt == dt->N * PTNSZ2) 
    dt->crt = 0; // restart
  int * ptn = dt->ptns + dt->crt;
  dt->crt += PTNSZ2;
  return ptn; 
}

int ptnN(struct data* dt) {
  if (dt->N == -1) {
    FILE* f = fopen(dt->ptnfile, "r");
    int n = 0;
    int tmpg, tmpm;
    while (fscanf(f, "%d%d", &tmpg, &tmpm) == 2) n++;
    fclose(f);
    dt->N = n;
  }
  return dt->N;
}

#define swapint(a, b) { int tmp; tmp = a; a = b; b = tmp; }

void rotate_board(int* ptn, int sym) {
  // sym == 0: unchanged
  if (sym == 1) {
    int i, j;
    for (i = 0; i < PTNSZ; i++)
      for (j = 0; j < PTNSZ/2; j++)
	swapint(ptn[i*PTNSZ + j], ptn[i*PTNSZ + PTNSZ - j - 1])
  }
  else if (sym == 2) {
    int i, j;
    for (j = 0; j < PTNSZ; j++)
      for (i = 0; i < PTNSZ/2; i++)
	swapint(ptn[i*PTNSZ + j], ptn[(PTNSZ - i - 1)*PTNSZ + j])
  }
  else  if (sym == 3) {
    int i, j;
    for (i = 0; i < PTNSZ; i++)
      for (j = 0; j < PTNSZ; j++) {
	int ij0 = i*PTNSZ + j, ij1 = (PTNSZ-i-1)*PTNSZ + (PTNSZ - j - 1);
	if (ij0 < ij1)
	  swapint(ptn[ij0], ptn[ij1]);
      }
  }
  else  if (sym == 4) {
    int i, j;
    for (i = 0; i < PTNSZ; i++)
      for (j = 0; j < i; j++)
	swapint(ptn[i*PTNSZ + j], ptn[j*PTNSZ + i])
  }
  else  if (sym == 5) {
    rotate_board(ptn, 4);
    rotate_board(ptn, 3);
  }
  else  if (sym == 6) {
    rotate_board(ptn, 4);
    rotate_board(ptn, 1);
  }
  else  if (sym == 7) {
    rotate_board(ptn, 4);
    rotate_board(ptn, 2);
  }
}

