%{
#include <stdio.h>
#include "neural_net_sgf-defs.h"
//#include "gamestate.h"
//#include "bookhash.h"

#define YYSTACKSIZE 100000

extern int yylex();
extern int yyparse();
extern int yyline;
extern char *yytext;
extern void yyerror(char* s);
extern void yyerrorf(char* fmt, ... );

void printBook(FILE* f, int moveno);

static char* move2sgf(int move, char *str);
static int sgf2move(char *str);

int rep_resmiss = 0;
int book_depth = 100;
int book_weight = 1;
#define USE_PASS
#define USE_SYM

int handicap_mode = 0;
%}

%union {
	int ptype;
        seqt seq;
}

%type <seq> sequence node properties

%token <ptype> AL_PROPID
%token AL_PAR_O AL_PAR_C AL_SC AL_PROPVAL

%%

collection:	gametree gametrees
gametrees:	gametree gametrees | ;
gametree:	AL_PAR_O sequence 
                gametrees AL_PAR_C
                { 
		  int i;
		  if (!handicap_mode) 
		    for (i = 0; i < $2.length; i++) {
		      if ($2.res != RE_STAT) 
			storeBookTT(hashLock_value(), hash_value(), 
				    book_weight, 
				    ($2.res==RE_BW)*book_weight, 
				    ($2.res==RE_WW)*book_weight);
		      undo_move();
		    }
		  else 
		    if ($2.res == RE_HA) {
		      handicap_mode = 0;
		      yyerror("back to normal");
		    }
		    else {
		      yyerrorf("handicap mode with result %d", $2.res);
		      exit(1);
		    }
		  
		  //  fprintf(stderr, "game-length: %d\n", $2.length); 
		  //  fprintf(stderr, "result: %d\n", $2.res);
		  //  fprintf(stderr, "eog: %d\n", yyline);
		  if (($2.res < 0) && rep_resmiss)
		    yyerror("missing result");
		}
sequence:	node sequence 
                { 
		  $$.length = $1.length + $2.length; 
		  if (handicap_mode && 
		      (($1.res == RE_HA) || ($2.res == RE_HA)))
		    $$.res = RE_HA;
		  else {
		    if ($1.res != RE_NONE) $$.res = $1.res;
		    else $$.res = $2.res;
		  }
		} 
		| { $$.length = 0; $$.res = RE_NONE; }
node:		AL_SC properties 
                { $$ = $2; 
		}
properties:	AL_PROPID AL_PROPVAL 
                { 
		  $$.res = RE_NONE;
		  $$.length = 0;
		  if ($1 == PT_HA) {
		    int ha;
		    if ((sscanf(yytext+1, "%d", &ha) == 1) && (ha>=2)){
		      yyerrorf("handicap game");
		      $$.res = RE_HA;
		      handicap_mode = 1;
		    }
		  }
		  if (!handicap_mode){
		    if ($1 == PT_RE) {
		      if (yytext[1] == 'B') $$.res = RE_BW; else
		      if (yytext[1] == 'W') $$.res = RE_WW; else
			$$.res = RE_EMPTY; 
		    }
		    if (($1 == PT_B) || ($1 == PT_W)) {
		      char isIllegal;
		      int move = sgf2move(yytext+1);
		      if (isOver())
			yyerrorf("game already over!");
		      else {
			if ((move == BOARDSURFACE) || isFree(move)) 
			  make_move(move, &isIllegal);
			else isIllegal = 1;
			if (isIllegal) {
			  yyerrorf("illegal move %s", yytext);
			  //  exit(1);
			}
			else $$.length = 1;
		      }
		    }
		    if ($1 == PT_C){
		      int gno, bwno, wwno;
		      if (sscanf(yytext+1, "stats: %d/%d/%d", 
				 &gno, &bwno, &wwno) == 3) {
			storeBookTT(hashLock_value(), hash_value(), 
				    gno*book_weight, 
				    bwno*book_weight, wwno*book_weight); 
			$$.res = RE_STAT;
		      }
		      else if (sscanf(yytext+1, "T: %d/%d/%d", 
				      &gno, &bwno, &wwno) == 3) 
			$$.res = RE_STAT;
		    }
		  } /* endif !handicap_mode */
		}
                propvals properties 
                {
		  if (handicap_mode) $$.length = 0; 
		  else $$.length = $3.length + $5.length; 
		  
		  if (handicap_mode && 
		      (($3.res == RE_HA) ||($5.res == RE_HA)))
		    $$.res = RE_HA;
		  else {
		      if ($3.res != RE_NONE) $$.res = $3.res; 
		      else $$.res = $5.res;
		    }
		}
	        | { $$.length = 0; $$.res = RE_NONE; }
propvals:	AL_PROPVAL propvals | ;

%%

#ifdef TEST_YY
extern FILE *yyin;


int main(int argc, char** argv){
  FILE * f;

  if (argc == 2) {
    yyin = fopen(argv[1], "r");
    if (yyin == NULL) {
      perror("");
      exit(1);
    }
  }

  init_boardstuff();
  clearBookTT();

  yyparse();

  fclose(yyin);

  f = fopen("book.sgf", "w");
  printBook(f, 0);
  fclose(f);

  printBookStats(stderr);
  return 0;
}

char* move2str(int move, char *str){
  int x, y;
  if (move == PASSMOVENR) { //Pass
    str[0] = 't'; str[1] = 't';
  }
  else {
    x = move % BOARDSIZE; y = move / BOARDSIZE;
    if (x > 7) x++;
    str[0] = x + 'a'; str[1] = y + '1'; str[2] = 0;
  }
  return str;
}

#endif

static char* move2sgf(int move, char *str){
  int x, y;
  if (move == PASSMOVENR) { 
    str[0] = 't'; str[1] = 't'; str[2] = 0;
  }
  else {
    x = move % BOARDSIZE; y = move / BOARDSIZE;
    str[0] = x + 'a'; str[1] = y + 'a'; str[2] = 0;
  }
  return str;
}

static int sgf2move(char *str){
  int x, y;
  if ((str[0] >= 't') || (str[0] < 'a')) return PASSMOVENR;
  x = str[0] - 'a';  
  y = str[1] - 'a';
  return x + BOARDSIZE*y;
}

#ifdef USE_PASS
#define MAXMOVES (BOARDSURFACE+1)
#else
#define MAXMOVES (BOARDSURFACE)
#endif
#ifdef USE_SYM
#define SYMNO 8
#else
#define SYMNO 1
#endif


void printBook(FILE* f, int moveno){
  int i, j, n, best;
  char isIllegal;
  int gno[MAXMOVES], bwno[MAXMOVES], wwno[MAXMOVES];
  char str[3];
  int btm = (moveno%2 == 0);

  int sym, symgno, symbwno, symwwno;
  unsigned int symhash[SYMNO];  
  long long unsigned int symhashLock[SYMNO];

  if (moveno == 0) fprintf(f, "( ;SZ[9]\n");

  n = 0;
  for (i = 0; i < MAXMOVES; i++){
    gno[i] = bwno[i] = wwno[i] = 0;
    if ((i == BOARDSURFACE) || isFree(i)) {
      make_move(i, &isIllegal);
      if (!isIllegal) {
	for (sym = 0; sym < SYMNO; sym++){
	  if (sym)
	    getSymHash(sym, &symhash[sym], &symhashLock[sym]);
	  else {
	    symhash[sym] = hash_value(); symhashLock[sym] = hashLock_value();
	  }
	  for (j = 0; j < sym; j++) 
	    if ((symhash[j] == symhash[sym]) && 
		(symhashLock[j] == symhashLock[sym])) break;
	  if (j == sym) 
	    if (retrieveBookTT(symhashLock[sym], symhash[sym], 
			       1, &symgno, &symbwno, &symwwno)){
	      gno[i] += symgno; 
	      bwno[i] += symbwno; 
	      wwno[i] += symwwno; 
	    }
	}
	undo_move();
      }
      if (gno[i]) n++;
    }
  }

  for (j = 0; j < n; j++){
    best = 0;

    for (i = 0; i < MAXMOVES; i++) 
      if (abs(gno[i]) > abs(gno[best])) best = i; 

    make_move(best, &isIllegal);
    if (n > 1) fprintf(f, "(");
    fprintf(f, ";%c[%s]", (btm?'B':'W'), move2sgf(best, str)); 
    if (gno[best] > 0) {
      fprintf(f, "C[stats: %d/%d/%d] ", gno[best], bwno[best], wwno[best]);
      if ((moveno < book_depth-1) && !isOver())
	printBook(f, moveno + 1);
    }
    else 
      fprintf(f, "C[T: %d/%d/%d] ", -gno[best], bwno[best], wwno[best]);
    if (n > 1) fprintf(f, ")\n");
    undo_move();

    gno[best] = 0;
  }

  if (moveno == 0) fprintf(f, ")\n");
}



