#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <math.h>

#include "network.h"
#include "data.h"

void trainnet(struct network* net, struct data* train, struct data* test, int T, int batchT, int testT, char* save);

#include <time.h>
inline void fprinttime(FILE* f, const char* s) {
  time_t tm; 
  time(&tm);
  fprintf(f, "#%s: %s", s, ctime(&tm));
}
#include <sys/resource.h>
void fprintmem(FILE* f, const char* s) {
  struct rusage usage;
  getrusage(RUSAGE_SELF, &usage);
  fprintf(f, "mem (%s): %ld\n", s, usage.ru_maxrss);
}
void fprintarg(FILE* f, int argc, char** argv) {
  fprintf(f, "#cmd:");
  int i;
  for (i = 0; i < argc; i++) {
    char *s = argv[i];
    for (; (*s) && (*s != ' '); s++);
    if (*s) fprintf(f, " \'%s\'", argv[i]);
    else fprintf(f, " %s", argv[i]);
  }
  fprintf(f, "\n");
  fprinttime(f, "start");
}


char *netfile = NULL;

double bmul = 1.0;

int unshareT = -1;

int main(int argc, char** argv) {
  fprintarg(stdout, argc, argv);
  fprintarg(stderr, argc, argv);
  // ptndir2 is for search-based features, suffix is .sb_ptn
  // this is specified with the "-a" option, other letters were taken!
  char *trainfile = NULL, *testfile = NULL, *ptndir = NULL, *ptndir2 = NULL;
  char *save = NULL;

  int T = 100, batchT = 1, testT = 10;

  int opt;
  while ((opt = getopt(argc, argv, "n:f:r:e:m:T:b:t:p:a:L:E:u:d:s:c:B:U:")) != -1) {
    switch (opt) {
    case 'n':
      netfile = strdup(optarg);
      break;
    case 'p':
      ptndir = strdup(optarg);
      break;
    case 'a':
      ptndir2 = strdup(optarg);
      break;
    case 'L':
      trainfile = strdup(optarg);
      break;
    case 'E':
      testfile = strdup(optarg);
      break;
    case 'f':
      featN = atoi(optarg);
      break;
    case 'r':
      RANDW = atof(optarg);
      break;
    case 'e':
      eta = atof(optarg);
      break;
    case 'c':
      etaC = atof(optarg);
    case 'm':
      mu = atof(optarg);
      break;
    case 'T':
      T = atoi(optarg);
      break;
    case 'b':
      batchT = atoi(optarg);
      break;
    case 't':
      testT = atoi(optarg);
      break;
    case 'u':
      updateRule = atoi(optarg);
      break;
    case 'd':
      wdecay = atof(optarg);
      break;
    case 's':
      save = strdup(optarg);
      break;
    case 'B':
      bmul = atof(optarg);
      break;
    case 'U':
      unshareT = atoi(optarg);
      break;
    default: /* '?' */
      fprintf(stderr, "Unknown option %c\n", opt);
      break;
    }
  }

  fprintmem(stderr, "start");
  struct network* net = create_net(netfile, 1);
  fprintf(stderr, "net created\n");
  fprintmem(stderr, "net");
  struct data* traindt = create_data(trainfile, ptndir, ptndir2, -1);
  struct data* testdt = create_data(testfile, ptndir, ptndir2, 1);
  fprintf(stderr, "data loaded\n");
  fprintmem(stderr, "data");

  trainnet(net, traindt, testdt, T, batchT, testT, save);

  fprinttime(stderr, "end");
  fprintmem(stderr, "end");
  return 0;
}

void fprintInput(FILE* f, int* input, int k0, int k1) {
  int i, m, k;
  for (k = k0, m = 0; k <= k1; k++) m = (m << 1) | 1; 
  fprintf(f, "\n");
  for (i = 0; i < PTNSZ2; i++) {
    fprintf(f, "%03x ", (input[i] >> k0) & m); 
    if ((i+1)%PTNSZ == 0) fprintf(f, "\n");
  }
  fprintf(f, "\n");
}

double testnet(struct network* net, struct data* test, int t1) {
  int testsz = ptnN(test);
  int t, i;
  double err[PTNSZ2];
  for (i = 0; i < PTNSZ2; i++) err[i] = 0;
  for (t = 0; t < testsz; t++) {
    struct sb_ptn *sb = nextptn(test);
    int* input = sb->ptn1;
    float* input2 = sb->ptn2;
    free(sb);
    //    fprintInput(stderr, input, 2, 4);
    double* output = response(net, input, input2);
    int best, n = 0;
    for (best = 0; best < PTNSZ2; best++) 
      if (input[best] & 1) break;
    if (isnan(output[best])) { 
	n = PTNSZ2-1;
    }
    else 
      for (i = 0; i < PTNSZ2; i++) 
	if ((i != best) && !(input[i] & 2) && (output[i] >= output[best]))
	  n++;
    for (i = 0; i < n; i++)
      err[i]++;
    //    fprintf(stderr, "test %d: %d %d %g %d\n", t1, t, best, output[best], n);
  }
  printf("%d ", t1);
  for (i = 0; i < PTNSZ2; i++)
    printf("%f ", err[i]/testsz);
  printf("#test\n");
  fflush(stdout);
  return err[0]/testsz;
}


void trainnet(struct network* net, struct data* train, struct data* test, 
	      int T, int batchT, int testT, char* save) {
  int t, batcht = batchT;
  double besterr = 1.0;
  double terr;

  if (unshareT == 0) unshareTop(net);
  
  testnet(net, test, 0);
  double err[PTNSZ2];
  int i;
  for (i = 0; i < PTNSZ2; i++) err[i] = 0;
  for (t = 1; t <= T; t++) {
    struct sb_ptn *input = nextptn(train);
    if (input == 0) break;
    int* ptn1 = input->ptn1;
    float* ptn2 = input->ptn2;
    free(input);
    //    fprintInput(stderr, input, 2, 4);
    double* output = response(net, ptn1, ptn2);
    double target[PTNSZ2];
    int valid[PTNSZ2];
    int n = 0, best = 0;
    for (i = 0; i < PTNSZ2; i++) {
      if (ptn1[i] & 1) best = i;
      valid[i] = !(ptn1[i] & 2);
    }
    for (i = 0; i < PTNSZ2; i++) target[i] = 0;
    target[best] = 1;
    
    if (isnan(output[best])) {
      n = PTNSZ2-1;
    }
    else
      for (i = 0; i < PTNSZ2; i++) 
	if ((i != best) && !(ptn1[i] & 2) && (output[i] >= output[best]))
	  n++;
    for (i = 0; i < n; i++)
      err[i]++;

    computegradient(net, ptn1, ptn2, target, valid, 0);

    batcht--;
    //    if (t % batchT == 0) {
    if (batcht == 0) {
      double *wsz = wsize(net);
      double *dwsz = dwsize(net);
      int l, L = layersN(net);
      fprintf(stderr, "wsz: %d ", t);
      for (l = 0; l < L; l++)
	fprintf(stderr, " %f", wsz[l]);
      fprintf(stderr, "\ndwsz: %d ", t);
      for (l = 0; l < L; l++)
	fprintf(stderr, " %f", dwsz[l]);
      fprintf(stderr, "\n");
      free(wsz); free(dwsz);

      update(net);
      fprintmem(stderr, "update");
      if (bmul >= 1)
	batchT *= bmul;
      else {
	if (pow(t, bmul) > batchT)
	  batchT = pow(t, bmul);
      }
	
      batcht = batchT;
      fprintf(stderr, "update %d %d\n", t, batchT);

      if (unshareT == t) unshareTop(net);
    }
    if (t % testT == 0) {
      printf("%d ", t);
      for (i = 0; i < PTNSZ2; i++) {
	printf("%f ", err[i]/testT);
	err[i] = 0;
      }
      printf("#train\n");
      fflush(stdout);

      terr = testnet(net, test, t);
      fflush(stdout);

      
      if ((save != NULL) && (terr < besterr)) {
	besterr = terr;
	fprintf(stderr, "saving new best (%f) at %d\n", besterr, t);
	print_net(save, net);
      }
    }
  }
}
